@@ -2,25 +2,25 @@
 
 from __future__ import annotations
 from typing import Union, Literal, Tuple, Optional, Dict
+import warnings
 import re
+from collections import ChainMap
 from math import inf, cos, sin, radians
 from itertools import zip_longest
 import importlib.resources as pkg_resources
 import xml.etree.ElementTree as ET
 
+from latex2mathml.converter import convert  # type: ignore
+import latex2mathml.commands  # type: ignore
+
 import ziafont as zf
 from ziafont.glyph import fmt
-
 from .mathfont import MathFont
-from .nodes import makenode, getstyle
+from .nodes import Mnode
+from .styles import parse_style
 from .escapes import unescape
 from .config import config
 
-try:
-    from latex2mathml.converter import convert  # type: ignore
-except ImportError:
-    convert = False
-
 
 Halign = Literal['left', 'center', 'right']
 Valign = Literal['top', 'center', 'base', 'axis', 'bottom']
@@ -37,13 +37,53 @@ def denamespace(element: ET.Element) -> ET.Element:
     return element
 
 
-def tex2mml(tex: str) -> str:
+def declareoperator(name: str) -> None:
+    r''' Declare a new operator name, similar to Latex ``\DeclareMathOperator`` command.
+
+        Args:
+            name: Name of operator, should start with a ``\``.
+                Example: ``declareoperator(r'\myfunc')``
+    '''
+    latex2mathml.commands.FUNCTIONS = latex2mathml.commands.FUNCTIONS + (name,)
+
+
+def tex2mml(tex: str, inline: bool = False) -> str:
     ''' Convert Latex to MathML. Do some hacky preprocessing to work around
         some issues with generated MathML that ziamath doesn't support yet.
     '''
     tex = re.sub(r'\\binom{(.+?)}{(.+?)}', r'\\left( \1 \\atop \2 \\right)', tex)
-    tex = re.sub(r'\\mathrm{(.+?)}', r'\\mathrm {\1}', tex)  # latex2mathml bug requires space after mathrm
-    return convert(tex)
+    # latex2mathml bug requires space after mathrm
+    tex = re.sub(r'\\mathrm{(.+?)}', r'\\mathrm {\1}', tex)
+    tex = tex.replace('||', '‖')
+    if config.decimal_separator == ',':
+        # Replace , with {,} to remove right space
+        # (must be surrounded by digits)
+        tex = re.sub(r'([0-9]),([0-9])', r'\1{,}\2', tex)
+
+    mml = convert(tex, display='inline' if inline else 'block')
+
+    # Replace some operators with "stretchy" variants
+    mml = re.sub(r'<mo>&#x0005E;', r'<mo>&#x00302;', mml)  # widehat
+    mml = re.sub(r'<mo>&#x0007E;', r'<mo>&#x00303;', mml)  # widetilde
+    return mml
+
+
+def apply_mstyle(element: ET.Element) -> ET.Element:
+    ''' Take attributes defined in <mstyle> elements and add them
+        to all the child elements, removing the original <mstyle>
+    '''
+    def flatten_attrib(element: ET.Element) -> None:
+        for child in element:
+            if element.tag == 'mstyle':
+                child.attrib = dict(ChainMap(child.attrib, element.attrib))
+            flatten_attrib(child)
+
+    flatten_attrib(element)
+
+    elmstr = ET.tostring(element).decode('utf-8')
+    elmstr = re.sub(r'<mstyle.+?>', '', elmstr)
+    elmstr = re.sub(r'</mstyle>', '', elmstr)
+    return ET.fromstring(elmstr)
 
 
 class Math:
@@ -54,34 +94,35 @@ class Math:
             size: Base font size, pixels
             font: Filename of font file. Must contain MATH typesetting table.
     '''
-    # Math class doesn't have a color parameter since color is set
-    # by the <mathcolor> tags in MML. Since Math is a single line,
-    # alignment is done at the drawon function rather than the class level.
     def __init__(self, mathml: Union[str, ET.Element],
-                 size: float=24, font: str=None):
-        self.mathml = mathml
-        self.size = size
+                 size: float = None, font: str = None):
+        self.size = size if size else config.math.fontsize
+        font = font if font else config.math.mathfont
 
+        self.font: MathFont
         if font is None:
-            self.font = loadedfonts.get('default')
+            self.font = loadedfonts['default']
         elif font in loadedfonts:
-            self.font = loadedfonts.get(font)
+            self.font = loadedfonts[font]
         else:
-            self.font = MathFont(font, size)
+            self.font = MathFont(font, self.size)
             loadedfonts[font] = self.font
 
         if isinstance(mathml, str):
             mathml = unescape(mathml)
             mathml = ET.fromstring(mathml)
         mathml = denamespace(mathml)
+        mathml = apply_mstyle(mathml)
 
-        self.style = getstyle(mathml)
+        self.mathml = mathml
+        self.style = parse_style(mathml)
         self.element = mathml
-        self.node = makenode(mathml, parent=self, size=size)  # type: ignore
+        self.mtag = 'math'
+        self.node = Mnode.fromelement(mathml, parent=self)  # type: ignore
 
     @classmethod
-    def fromlatex(cls, latex: str, size: float=24, mathstyle: str=None,
-                  font: str=None, color: str=None):
+    def fromlatex(cls, latex: str, size: float = None, mathstyle: str = None,
+                  font: str = None, color: str = None, inline: bool = False):
         ''' Create Math Renderer from a single LaTeX expression. Requires
             latex2mathml Python package.
 
@@ -91,12 +132,10 @@ class Math:
                 mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
                 font: Font file name
                 color: Color parameter, equivalent to "mathcolor" attribute
+                inline: Use inline math mode (default is block mode)
         '''
-        if not convert:
-            raise ValueError('fromlatex requires latex2mathml package.')
-
         mathml: Union[str, ET.Element]
-        mathml = tex2mml(latex)
+        mathml = tex2mml(latex, inline=inline)
         if mathstyle:
             mathml = ET.fromstring(mathml)
             mathml.attrib['mathvariant'] = mathstyle
@@ -107,9 +146,9 @@ class Math:
         return cls(mathml, size, font)
 
     @classmethod
-    def fromlatextext(cls, latex: str, size: float=24, mathstyle: str=None,
-                      textstyle: str=None, font: str=None,
-                      color: str=None):
+    def fromlatextext(cls, latex: str, size: float = 24, mathstyle: str = None,
+                      textstyle: str = None, font: str = None,
+                      color: str = None):
         ''' Create Math Renderer from a sentence containing zero or more LaTeX
             expressions delimited by $..$, resulting in single MathML element.
             Requires latex2mathml Python package.
@@ -122,12 +161,15 @@ class Math:
                 font: Font file name
                 color: Color parameter, equivalent to "mathcolor" attribute
         '''
+        warnings.warn(r'fromlatextext is deprecated. Use ziamath.Text or \text{} command.', DeprecationWarning, stacklevel=2)
+        
         # Extract each $..$, convert to MathML, but the raw text in <mtext>, and join
         # into a single <math>
-        parts = re.split(r'\$(.*?)\$', latex)
+        parts = re.split(r'(\$+.*?\$+)', latex)
         texts = parts[::2]
-        maths = [tex2mml(p) for p in parts[1::2]]
-        mathels = [ET.fromstring(m)[0] for m in maths]   # Convert to xml, but drop opening <math>
+        maths = [tex2mml(p.replace('$', ''), inline=not p.startswith('$$')) for p in parts[1::2]]
+        mathels = [ET.fromstring(m) for m in maths]   # Convert to xml, but drop opening <math>
+
         mml = ET.Element('math')
         for text, mathel in zip_longest(texts, mathels):
             if text:
@@ -136,9 +178,13 @@ class Math:
                     mtext.attrib['mathvariant'] = textstyle
                 mtext.text = text
             if mathel is not None:
+                child = mathel[0]
                 if mathstyle:
-                    mathel.attrib['mathvariant'] = mathstyle
-                mml.append(mathel)
+                    child.attrib['mathvariant'] = mathstyle
+                if (dstyle := mathel.attrib.get('display')):
+                    child.attrib['displaystyle'] = {'inline': 'false',
+                                                    'block': 'true'}.get(dstyle, 'true')
+                mml.append(child)
         if color:
             mml.attrib['mathcolor'] = color
         return cls(mml, size, font)
@@ -160,8 +206,8 @@ class Math:
         svg.attrib['viewBox'] = f'{fmt(bbox.xmin-1)} {fmt(-bbox.ymax-1)} {fmt(width)} {fmt(height)}'
         return svg
 
-    def drawon(self, svg: ET.Element, x: float=0, y: float=0,
-               halign: Halign='left', valign: Valign='base') -> ET.Element:
+    def drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
+               halign: Halign = 'left', valign: Valign = 'base') -> ET.Element:
         ''' Draw the math expression on an existing SVG
 
             Args:
@@ -180,7 +226,7 @@ class Math:
         width, height = self.getsize()
         yshift = {'top': self.node.bbox.ymax,
                   'center': height/2 + self.node.bbox.ymin,
-                  'axis': self.font.math.consts.axisHeight * self.node.emscale,  # type: ignore
+                  'axis': self.node.units_to_points(self.font.math.consts.axisHeight),
                   'bottom': self.node.bbox.ymin}.get(valign, 0)
         xshift = {'center': -width/2,
                   'right': -width}.get(halign, 0)
@@ -204,7 +250,7 @@ class Math:
 
     @classmethod
     def mathml2svg(cls, mathml: Union[str, ET.Element],
-                   size: float=24, font: str=None):
+                   size: float = None, font: str = None):
         ''' Shortcut to just return SVG string directly '''
         return cls(mathml, size=size, font=font).svg()
 
@@ -218,9 +264,37 @@ class Math:
         return self.node.bbox.ymin
 
 
+class Latex(Math):
+    ''' Render Math from LaTeX
+
+        Args:
+            latex: Latex string
+            size: Base font size
+            mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
+            font: Font file name
+            color: Color parameter, equivalent to "mathcolor" attribute
+            inline: Use inline math mode (default is block mode)
+        '''
+    def __init__(self, latex: str, size: float = None, mathstyle: str = None,
+                 font: str = None, color: str = None, inline: bool = False):
+        self.latex = latex
+
+        mathml: Union[str, ET.Element]
+        mathml = tex2mml(latex, inline=inline)
+        if mathstyle:
+            mathml = ET.fromstring(mathml)
+            mathml.attrib['mathvariant'] = mathstyle
+            mathml = ET.tostring(mathml, encoding='unicode')
+        if color:
+            mathml = ET.fromstring(mathml)
+            mathml.attrib['mathcolor'] = color
+        super().__init__(mathml, size, font)
+
+
 class Text:
-    ''' Mixed text and latex math, with math delimited by $..$. Drawn to SVG.
-        Can contain multiple lines.
+    ''' Mixed text and latex math. Inline math delimited by single $..$, and
+        display-mode math delimited by double $$...$$. Can contain multiple
+        lines. Drawn to SVG.
 
         Args:
             s: string to write
@@ -238,17 +312,18 @@ class Text:
                 https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
 
     '''
-    def __init__(self, s, textfont: str=None, mathfont: str=None,
-                 mathstyle: str=None, size: float=24, linespacing: float=1,
-                 color: str=None,
-                 halign: str='left', valign: str='base',
-                 rotation: float=0, rotation_mode: str='anchor'):
+    def __init__(self, s, textfont: str = None, mathfont: str = None,
+                 mathstyle: str = None, size: float = None, linespacing: float = None,
+                 color: str = None,
+                 halign: str = 'left', valign: str = 'base',
+                 rotation: float = 0, rotation_mode: str = 'anchor'):
         self.str = s
         self.mathfont = mathfont
         self.mathstyle = mathstyle
-        self.size = size
-        self.linespacing = linespacing
+        self.size = size if size else config.text.fontsize
+        self.linespacing = linespacing if linespacing else config.text.linespacing
         self.color = color
+        self.textcolor = color if color else config.text.color
         self._halign = halign
         self._valign = valign
         self.rotation = rotation
@@ -257,14 +332,18 @@ class Text:
 
         # textfont can be a path to font, or style type like "serif".
         # If style type, use Stix font variation
+        textfont = textfont if textfont else config.text.textfont
+        textstyle = config.text.variant
         if textfont is None:
-            textfont = 'sans'
-        if textfont in ['sans', 'sans-serif', 'serif']:
             self.textfont = None
-            self.textstyle = textfont
+            self.textstyle = textstyle
         else:
-            self.textfont = zf.Font(textfont)
-            self.textstyle = 'sans'
+            try:
+                self.textfont = zf.Font(textfont)
+                self.textstyle = textstyle
+            except FileNotFoundError:
+                self.textfont = None
+                self.textstyle = textfont
 
     def svg(self) -> str:
         ''' Get expression as SVG string '''
@@ -277,7 +356,7 @@ class Text:
     def svgxml(self) -> ET.Element:
         ''' Get standalone SVG of expression as XML Element Tree '''
         svg = ET.Element('svg')
-        svgelm, (x1, x2, y1, y2) = self._drawon(svg)
+        _, (x1, x2, y1, y2) = self._drawon(svg)
         svg.attrib['width'] = fmt(x2-x1)
         svg.attrib['height'] = fmt(y2-y1)
         svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
@@ -291,8 +370,8 @@ class Text:
         with open(fname, 'w') as f:
             f.write(self.svg())
 
-    def drawon(self, svg: ET.Element, x: float=0, y: float=0,
-               halign: str=None, valign: str=None) -> ET.Element:
+    def drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
+               halign: str = None, valign: str = None) -> ET.Element:
         ''' Draw text on existing SVG element
 
             Args:
@@ -305,8 +384,8 @@ class Text:
         svgelm, _ = self._drawon(svg, x, y, halign, valign)
         return svgelm
 
-    def _drawon(self, svg: ET.Element, x: float=0, y: float=0,
-                halign: str=None, valign: str=None) -> Tuple[ET.Element, Tuple[float, float, float, float]]:
+    def _drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
+                halign: str = None, valign: str = None) -> Tuple[ET.Element, Tuple[float, float, float, float]]:
         ''' Draw text on existing SVG element
 
             Args:
@@ -327,24 +406,40 @@ class Text:
         linesizes = []
         for line in lines:
             svgparts = []
-            parts = re.split(r'(\$.*?\$)', line)
+            parts = re.split(r'(\$+.*?\$+)', line)
             partsizes = []
             for part in parts:
                 if not part:
                     continue
-                if part.startswith('$') and part.endswith('$'):  # Math
-                    math = Math.fromlatex(part.replace('$', ''), font=self.mathfont,
+                if part.startswith('$$') and part.endswith('$$'):  # Display-mode math
+                    math = Math.fromlatex(part.replace('$', ''),
+                                          font=self.mathfont,
                                           mathstyle=self.mathstyle,
+                                          inline=False,
+                                          size=self.size, color=self.color)
+                    svgparts.append(math)
+                    partsizes.append(math.getsize())
+
+                elif part.startswith('$') and part.endswith('$'):  # Text-mode Math
+                    math = Math.fromlatex(part.replace('$', ''),
+                                          font=self.mathfont,
+                                          mathstyle=self.mathstyle,
+                                          inline=True,
                                           size=self.size, color=self.color)
                     svgparts.append(math)
                     partsizes.append(math.getsize())
                 else:  # Text
                     if self.textfont:
-                        txt = zf.Text(part, font=self.textfont, size=self.size, color=self.color)
+                        # A specific font file is defined, use ziafont and ignore textstyle
+                        txt = zf.Text(part, font=self.textfont, size=self.size, color=self.textcolor)
                         partsizes.append(txt.getsize())
                     else:
-                        txt = Math.fromlatextext(part, textstyle=self.textstyle,
-                                                 size=self.size, color=self.color)
+                        # use math font with textstyle
+                        txt = Math.fromlatex(fr'\text{{{part}}}',
+                                             font=self.mathfont,
+                                             mathstyle=self.textstyle,
+                                             size=self.size,
+                                             color=self.textcolor)
                         partsizes.append((txt.node.bbox.xmax - txt.node.bbox.xmin,  # type: ignore
                                           txt.node.size))  # type: ignore
                     svgparts.append(txt)
@@ -425,7 +520,7 @@ class Text:
                         bbox[2]+dy, bbox[3]+dy)
 
             xform += f' rotate({-self.rotation} {x} {y})'
-            if config.debug:
+            if config.debug.bbox:
                 rect = ET.SubElement(svg, 'rect')
                 rect.attrib['x'] = fmt(bbox[0])
                 rect.attrib['y'] = fmt(bbox[2])
@@ -455,4 +550,4 @@ class Text:
 # Cache the loaded fonts to prevent reloading all the time
 with pkg_resources.path('ziamath.fonts', 'STIXTwoMath-Regular.ttf') as p:
     fontname = p
-loadedfonts: Dict[str, Union[zf.Font, MathFont]] = {'default': MathFont(fontname)}
+loadedfonts: Dict[str, MathFont] = {'default': MathFont(fontname)}
